﻿/*
Direct3D Device Wrapper Class
20 July 2008
*/

#include <windows.h>
#include <d3d9.h>
#pragma hdrstop

#include "tlib.h"

#include "Util.h"

#include "CVector.h"
#include "CMatrix.h"
#include "CVertexColour.h"
#include "CVertexTexture.h"
#include "CDevice.h"
#include "CTexture.h"
#include "CSprite.h"

using namespace tlib;
using namespace LithDX;



CDevice::CDevice(IDirect3DDevice9* aDevice)
		: cVertexColourFormat(D3DFVF_XYZ | D3DFVF_DIFFUSE),
		cVertexTextureFormat(D3DFVF_XYZ | D3DFVF_TEX1)
{
	mDevice = aDevice;

	mBackgroundColour = D3DCOLOR_XRGB(0, 0, 0);

	mSpriteViewportXSize = 1024;
	mSpriteViewportYSize = 768;
}



CDevice::~CDevice()
{
	for (int i = 0; i < mSprites.NumItems; i++)
		delete mSprites[i];
	for (int i = 0; i < mTextures.NumItems; i++)
		delete mTextures[i];
	mDevice->Release();
}



void CDevice::SetTransform(D3DTRANSFORMSTATETYPE aState, const CMatrix& aMatrix)
{
	D3DMATRIX Transform;
	for (int y = 0; y < 4; y++)
	{
		for (int x = 0; x < 4; x++)
			Transform.m[y][x] = aMatrix.DataC(x, y);
	}
	CheckHResult(mDevice->SetTransform(aState, &Transform));
}



CTexture* CDevice::CreateTexture(int aXSize, int aYSize, const TByte* aData)
{
	IDirect3DTexture9 *TempTexture;
	CheckHResult(mDevice->CreateTexture(aXSize, aYSize, 1, 0, D3DFMT_A8R8G8B8,
			D3DPOOL_SYSTEMMEM, &TempTexture, NULL));
	D3DLOCKED_RECT Rect;
	CheckHResult(TempTexture->LockRect(0, &Rect, NULL, 0));
	memcpy(Rect.pBits, aData, aXSize * aYSize * 4);
	CheckHResult(TempTexture->UnlockRect(0));
	IDirect3DTexture9 *Texture;
	CheckHResult(mDevice->CreateTexture(aXSize, aYSize, 1, 0, D3DFMT_A8R8G8B8,
			D3DPOOL_DEFAULT, &Texture, NULL));
	CheckHResult(mDevice->UpdateTexture(TempTexture, Texture));
	TempTexture->Release();

	CTexture *WrappedTexture = new CTexture(Texture);
	mTextures.Add(WrappedTexture);
	return WrappedTexture;
}



CSprite* CDevice::CreateSprite(int aXSize, int aYSize, const TByte* aData)
{
	CTexture *Texture = CreateTexture(aXSize, aYSize, aData);
	CSprite *Sprite = new CSprite(aXSize, aYSize, Texture);
	mSprites.Add(Sprite);
	return Sprite;
}



void CDevice::BeginFrame()
{
	CheckHResult(mDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, mBackgroundColour,
			1.0, 0));
	CheckHResult(mDevice->BeginScene());
}



void CDevice::EndFrame()
{
	CheckHResult(mDevice->EndScene());
	CheckHResult(mDevice->Present(NULL, NULL, NULL, NULL));
}



void CDevice::TriColour(const CVertexColour& aV0, const CVertexColour& aV1,
		const CVertexColour& aV2)
{
	static CVertexColour Vertex[3];
	Vertex[0] = aV0;
	Vertex[1] = aV1;
	Vertex[2] = aV2;

	CheckHResult(mDevice->SetFVF(cVertexColourFormat));
	CheckHResult(mDevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, 1, Vertex, sizeof(CVertexColour)));
}



void CDevice::TriTexture(const CVertexTexture& aV0, const CVertexTexture& aV1,
		const CVertexTexture& aV2, CTexture* aTexture)
{
	static CVertexTexture Vertex[3];
	Vertex[0] = aV0;
	Vertex[1] = aV1;
	Vertex[2] = aV2;

	CheckHResult(mDevice->SetFVF(cVertexTextureFormat));
	CheckHResult(mDevice->SetTexture(0, aTexture->GetTexture()));
	CheckHResult(mDevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, 1, Vertex, sizeof(CVertexTexture)));
	CheckHResult(mDevice->SetTexture(0, NULL));
}



void CDevice::SetSpriteViewport(int aXSize, int aYSize)
{
	mSpriteViewportXSize = aXSize;
	mSpriteViewportYSize = aYSize;
}



void CDevice::DrawSprite(CSprite* aSprite, int aX, int aY, double aXAlign, double aYAlign)
{
	double X0 = ((double)aX - aXAlign * aSprite->GetXSize());
	double X1 = X0 + aSprite->GetXSize();
	double Y0 = ((double)aY - aYAlign * aSprite->GetYSize());
	double Y1 = Y0 + aSprite->GetYSize();

	float VX0 = -1 + (X0 / mSpriteViewportXSize * 2);
	float VY0 = 1 - (Y0 / mSpriteViewportYSize * 2);
	float VX1 = -1 + (X1 / mSpriteViewportXSize * 2);
	float VY1 = 1 - (Y1 / mSpriteViewportYSize * 2);

	CVertexTexture V0(CVector(VX0, VY0, 0), 0, 0);
	CVertexTexture V1(CVector(VX1, VY0, 0), 1, 0);
	CVertexTexture V2(CVector(VX0, VY1, 0), 0, 1);
	CVertexTexture V3(CVector(VX1, VY1, 0), 1, 1);
	TriTexture(V0, V1, V2, aSprite->GetTexture());
	TriTexture(V1, V3, V2, aSprite->GetTexture());
}