﻿#include <windows.h>
#pragma hdrstop

#include "tlib.h"

#include "CFile.h"

using namespace tlib;
using namespace LithUtil;



CFile::CFile(TString aFilePath, EOpenMode aOpenMode, bool aReadSharing)
{
	switch (aOpenMode)
	{
	case omCreate:
		mFile = CreateFile(aFilePath.Str, GENERIC_WRITE, aReadSharing ? FILE_SHARE_READ : 0,
				NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case omRead:
		mFile = CreateFile(aFilePath.Str, GENERIC_READ, aReadSharing ? FILE_SHARE_READ : 0,
				NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case omAppend:
		mFile = CreateFile(aFilePath.Str, GENERIC_WRITE, aReadSharing ? FILE_SHARE_READ : 0,
				NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	}
	if (mFile == INVALID_HANDLE_VALUE)
		throw TException(L"Failed to open file");

	if (aOpenMode == omAppend)
		Seek(0, smEnd);

	mEOF = false;
}



CFile::~CFile()
{
	/*
	if (FlushFileBuffers(mFile) == 0)
		throw TException(L"FlushFileBuffers() failed");
	*/
	if (CloseHandle(mFile) == 0)
		throw TException(L"CloseHandle() failed");
}



/**
Reads the specified number of bytes into the specified buffer.
Returns the actual number of bytes read.
*/
unsigned int CFile::Read(TByte* aBuffer, unsigned int aSize)
{
	DWORD BytesRead;
	if (ReadFile(mFile, aBuffer, aSize, &BytesRead, NULL) == 0)
		throw TException(L"ReadFile() failed");
	if (BytesRead < aSize)
		mEOF = true;
	return BytesRead;
}



int CFile::ReadInt()
{
	int Value;
	if (Read((TByte*)&Value, sizeof(int)) < sizeof(int))
		throw TException(L"Failed to read integer");
	return Value;
}



/**
Reads a unicode string from the data. The string data is preceeded by an integer containing
the number of UTF-16 elements in the string.
*/
TString CFile::ReadString()
{
	int StringSize = ReadInt();

	wchar_t *Data = new wchar_t [StringSize];
	if (Read((TByte*)Data, StringSize * sizeof(wchar_t)) < StringSize)
		throw TException(L"Failed to read string data");
	TString Result(StringSize, Data);
	delete [] Data;
	return Result;
}



/**
Reads a single unicode character from the data.
Returns true iff the character was invalid.
Returns false if there is no more data left. Check for IsEOF to detect this case.
*/
bool CFile::ReadCharacter(int& aCharacter)
{
	wchar_t e;
	unsigned int BytesRead = Read((TByte*)&e, sizeof(wchar_t));
	if (BytesRead == 0)
	{
		aCharacter = 0;
		return false;
	}
	if (BytesRead < sizeof(wchar_t))
		return true;
	if (e >= 0xdc00 && e < 0xe000)
		return true;
	if (!(e >= 0xd800 && e < 0xdc00))
	{
		aCharacter = (int)e;
		return false;
	}

	// read surrogate pair
	wchar_t e1;
	if (Read((TByte*)&e1, sizeof(wchar_t)) < sizeof(wchar_t))
		return true;
	if (!(e1 >= 0xdc00 && e1 < 0xe000))
		return true;
	aCharacter = ((e & 0x3ff) << 10l | (e1 & 0x3ff)) + 0x10000;
	return false;
}



/**
Reads a line of text from a unicode text file.
Returns true iff the data contains invalid text data.
*/
bool CFile::ReadLine(TString& aLine)
{
	TStringBuilder Line;
	while (true)
	{
		int c;
		if (ReadCharacter(c))
			return true;
		if (mEOF || c == L'\r' || c == L'\n')
		{
			if (c == L'\r')
				Seek(1, smCurrent);
			aLine = Line.GetString();
			return false;
		}
		Line.AppendChar(c);
	}
}



void CFile::Write(const TByte* aBuffer, unsigned int aSize)
{
	DWORD BytesWritten;
	if (WriteFile(mFile, aBuffer, aSize, &BytesWritten, NULL) == 0)
		throw TException(L"WriteFile() failed");
}



void CFile::WriteInt(int aValue)
{
	Write((TByte*)&aValue, sizeof(int));
}



void CFile::WriteString(TString aValue)
{
	WriteInt(aValue.Size);
	Write((const TByte*)aValue.Str, aValue.Size * sizeof(wchar_t));
}



void CFile::WriteText(TString aText)
{
	Write((const TByte*)aText.Str, aText.Size * sizeof(wchar_t));
}



void CFile::Seek(unsigned int aOffset, ESeekMode aSeekMode)
{
	if (SetFilePointer(mFile, aOffset, NULL, aSeekMode) == INVALID_SET_FILE_POINTER)
		throw TException(L"SetFilePointer() failed");
}



unsigned int CFile::GetCurrentPosition()
{
	DWORD Position = SetFilePointer(mFile, 0, NULL, FILE_CURRENT);
	if (Position == INVALID_SET_FILE_POINTER)
		throw TException(L"SetFilePointer() failed");
	return Position;
}



void CFile::Flush()
{
	if (FlushFileBuffers(mFile) == 0)
		throw TException(L"FlushFileBuffers() failed");
}



unsigned int CFile::GetSize()
{
	DWORD Size = GetFileSize(mFile, NULL);
	if (Size == INVALID_FILE_SIZE)
		throw TException(L"GetFileSize() failed");
	return Size;
}